//
// LRat.java
//
// Bill Seymour, 2025-03-17
//
// Copyright 2025 Bill Seymour.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This is an open-source rational number in Java that just uses longs
// for its numerator and denominator.
//
// The existing Rational.java uses BigIntegers instead of longs internally
// to help guard against overflow.  If you want to use this version instead,
// you'll need to rename the two source files.
//

import java.math.RoundingMode;

public final class Rational extends Number implements Comparable<Rational>
{
    private long num, den;

    public static final Rational ZERO = new Rational();
    public static final Rational ONE = new Rational(1L);
    public static final Rational ONE_HALF = new Rational(1L, 2L);

    public Rational()
    {
        num = 0L;
        den = 1L;
    }

    public Rational(long val)
    {
        num = val;
        den = 1L;
    }

    //
    // Greatest common divisor, used only by norm(long,long):
    //
    // Requires a != 0 && c > 0
    //
    private static long gcd(long a, long c)
    {
        if (a < 0L)
        {
            a = -a;
        }
        //
        // Euclid's algorithm (Knuth, vol. 2, sect. 4.5.2):
        //
        do {
            long r = a % c;
            a = c;
            c = r;
        } while (c != 0L);
        return a;
    }
    //
    // The following private constructor helper normalizes the fraction
    // such that the denominator is greater than zero and the fraction
    // is reduced to lowest terms.  It does all of Rational(long,long)'s
    // work and is shared by Rational(String,int).
    //
    private void norm(long n, long d)
    {
        if (d == 0L)
        {
            throw new ArithmeticException(
                "Constructing rational with zero denominator");
        }
        if (n == 0L)
        {
            num = 0L;
            den = 1L;
        }
        else
        {
            if (d < 0L)
            {
                n = -n;
                d = -d;
            }
            long g = gcd(n, d);
            if (g != 1L)
            {
                n /= g;
                d /= g;
            }
            num = n;
            den = d;
        }
    }
    public Rational(long n, long d)
    {
        norm(n, d);
    }

    //
    // Exact conversion of double by twidling bits:
    //
    // Assume IEEE 754 doubles:  52 significand bits,
    // 11 exponent bits excess 1023, and a sign.
    //
    // This is the version that just uses longs for the numerator
    // and denominator, so we need to worry about overflow.
    //
    // First, avoid a bunch of magic numbers:
    //
    private static final long SIGN_BIT = 0x8000_0000_0000_0000L;
    private static final long EXP_BITS = 0x7ff0_0000_0000_0000L;
    private static final long SIG_BITS = 0x000f_ffff_ffff_ffffL;
    private static final long NORM_MSB = 0x0010_0000_0000_0000L;
    private static final int EXP_BIT_COUNT = 11;
    private static final int SIG_BIT_COUNT = 52;
    private static final int EXP_OFFSET = 1075; // excess 1023 + 52 sig. bits
                                                // so that sig. LSB is 2**exp
    public Rational(double val)
    {
        //
        // Initialize num and den (we'll change the values later):
        //
        num = Double.doubleToRawLongBits(val); // all bits for now
        den = 1L; // adjusted later if we have a negative exponent

        if ((num & ~SIGN_BIT) == 0L) // the value zero
        {
            num = 0L; // just in case val is a negative zero
            return;   // NB:  early return
        }

        int exp = (int)((num & EXP_BITS) >> SIG_BIT_COUNT);
        if (exp == 0x7ff)
        {
            throw new IllegalArgumentException(
              "Can't construct rational from non-finite double");
        }
        if (exp == 0)
        {
            throw new ArithmeticException(
              "Overflow constructing rational from subnormal double");
        }

        //
        // At this point, we know that the value is finite, normal,
        // and non-zero; so there are no more special-case bit patterns.
        //
        boolean isNeg = num < 0L;
        num &= SIG_BITS;   // now just the significand bits
        num |= NORM_MSB;   // assumed MSB for normals
        exp -= EXP_OFFSET; // excess (1023 + 52 bits), so LSB now 2**exp

        //
        // If the exponent is less than zero, we'll multiply the denominator
        // by some power of two; so to make sure that we're in lowest terms,
        // discard trailing zeros until either the current numerator is odd
        // or the exponent is no longer negative:
        //
        while (exp < 0 && (num & 1L) == 0L)
        {
            num >>= 1;
            ++exp;
        }

        //
        // Adjust the value given the exponent:
        //
        if (exp > 0)
        {
            if (exp >= EXP_BIT_COUNT)
            {
                //
                // The assumed MSB for normals would shift into the sign bit.
                //
                throw new ArithmeticException(
                    "Numerator overflows constructing rational from double");
            }
            num <<= exp;
        }
        else if (exp < 0)
        {
            if (exp <= -(Long.SIZE - 1))
            {
                //
                // The current denominator, 1L, would shift into the sign bit.
                //
                throw new ArithmeticException(
                    "Denominator overflows constructing rational from double");
            }
            den <<= -exp;
        }

        //
        // Since either den is 1 or num is odd and den is a power of 2,
        // we know that we're in lowest terms; so all that remains is:
        //
        if (isNeg)
        {
            num = -num;
        }
    }

    //
    // Approximate conversion of double by continued fractions:
    //
    public Rational(double val, double err)
    {
        if (!Double.isFinite(val))
        {
            throw new IllegalArgumentException(
                "Can't construct Rational from non-finite double");
        }
        if (Math.abs(val) > Long.MAX_VALUE)
        {
            throw new ArithmeticException(
                "Overflow constructing rational from double");
        }

        boolean isNeg = val < 0.0;
        if (isNeg)
        {
            val = -val;
        }

        if (val == 0.0)
        {
            num = 0L;
            den = 1L;
            return; // NB:  early return
        }

        double maxError = Math.abs(val * err);
        if (!Double.isFinite(maxError) || maxError < Double.MIN_NORMAL)
        {
            throw new ArithmeticException(
                "Non-finite or subnormal maximum error " +
                "when constructing Rational from double");
        }

        // continued fractions (Knuth, vol. 2, sect. 4.5.3):

        double val0 = val; // current fraction term

        long int0 = (long)val0;

        long num0 = int0, den0 = 1L; // first guess:  (long)val / 1

        long num1 = 1L, num2; // prev. and 2nd-prev. numerator
        long den1 = 0L, den2; // prev. and 2nd-prev. denominator

        double err0 = Math.abs(val0 - (double)int0);

        while (err0 > maxError)
        {
            val0 = 1.0 / (val0 - (double)int0);
            if (!Double.isFinite(val0))
            {
                break; // the best we can do
            }
            int0 = (long)val0;

            num2 = num1;
            num1 = num0;
            num0 = int0 * num1 + num2;

            den2 = den1;
            den1 = den0;
            den0 = int0 * den1 + den2;

            double err1 = err0;
            err0 = Math.abs(val - (double)num0 / (double)den0);

            if (err0 >= err1) // no longer converging
            {
                break; // the best we can do
            }
        }

        //
        // Continued fractions result is known to be in lowest terms already.
        //
        num = isNeg ? -num0 : num0;
        den = den0;
    }

    public Rational(String val, int radix)
    {
        int punctuator = val.indexOf('/');
        if (punctuator >= 0)
        {
            // numerator/denominator
            norm(Long.parseLong(val.substring(0, punctuator), radix),
                 Long.parseLong(val.substring(punctuator + 1), radix));
        }
        else // no slash
        {
            punctuator = val.indexOf('.');
            if (punctuator < 0)
            {
                // just an integer
                num = Long.parseLong(val, radix);
                den = 1L;
            }
            else
            {
                // a single value with a radix point
                StringBuilder raw = new StringBuilder(val);
                raw.deleteCharAt(punctuator);

                // unscaled value / radix**scale
                long newden = 1L;
                for (int scale = raw.length() - punctuator; scale > 0; --scale)
                {
                    newden *= radix;
                }
                norm(Long.parseLong(raw.toString(), radix), newden);
            }
        }
    }
    public Rational(String val)
    {
        this(val, 10);
    }

    @Override public int compareTo(Rational rhs)
    {
        long diff = this.num * rhs.den - this.den * rhs.num;
        return diff < 0L ? -1 : (diff > 0L ? +1 : 0);
    }
    @Override public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        if (!(obj instanceof Rational))
        {
            return false;
        }
        Rational rhs = (Rational)obj;
        // We know we're in lowest terms.
        return this.num == rhs.num && this.den == rhs.den;
    }
    @Override public int hashCode()
    {
        return Long.hashCode(num) ^ Long.hashCode(den);
    }

    public long numerator()
    {
        return num;
    }
    public long denominator()
    {
        return den;
    }
    public int signum()
    {
        return num < 0L ? -1 : (num > 0L ? +1 : 0);
    }
    public Rational[] getParts()
    {
        Rational intPart  = new Rational(num / den); // rounds toward 0
        Rational fracPart = subtract(intPart);
        return new Rational[] { intPart, fracPart };
    }

    public Rational round(RoundingMode rnd)
    {
        if (den == 1L)
        {
            return this; // It's already an integer.
        }

        //
        // Because we know we're in lowest terms, we can now infer that
        // the numerator is non-zero and is not divisible by the denominator.
        //
        long intPart = Math.abs(num) / den;
        //
        // intPart is now the absolute value of this rounded toward zero.
        //

        boolean isNeg = num < 0; // We'll change the sign later.

        //
        // Other flags that we might not need to compute:
        //
        boolean isHalf = false;
        boolean isLess = false;
        boolean isOdd  = false;
        if (rnd == RoundingMode.HALF_UP   ||
            rnd == RoundingMode.HALF_DOWN ||
            rnd == RoundingMode.HALF_EVEN)
        {
            int halfness =
                this.abs().subtract(new Rational(intPart)).compareTo(ONE_HALF);
            isHalf = halfness == 0;
            isLess = halfness < 0;
            isOdd  = (intPart & 1L) != 0L;
        }

        switch (rnd)
        {
            case UP:
                ++intPart;
                break;
            case DOWN:
                // already done
                break;

            case CEILING:
                if (!isNeg)
                {
                    ++intPart;
                }
                break;
            case FLOOR:
                if (isNeg)
                {
                    ++intPart;
                }
                break;

            case HALF_UP:
                if (!isLess)
                {
                    ++intPart;
                }
                break;
            case HALF_DOWN:
                if (!isHalf && !isLess)
                {
                    ++intPart;
                }
                break;
            case HALF_EVEN:
                if (isOdd && !isLess)
                {
                    ++intPart;
                }
                break;

            case UNNECESSARY:
                throw new ArithmeticException("Rational.round(UNNECESSARY)");
            default:
                throw new IllegalArgumentException("Rational.round(garbage)");
        }

        return new Rational(isNeg ? -intPart : intPart);
    }
    public Rational round()
    {
        return round(RoundingMode.HALF_UP);
    }
    public Rational ceil()
    {
        return round(RoundingMode.CEILING);
    }
    public Rational floor()
    {
        return round(RoundingMode.FLOOR);
    }
    public Rational rint()
    {
        return round(RoundingMode.HALF_EVEN);
    }
    public Rational trunc()
    {
        // shorter RoundingMode.DOWN:
        return new Rational(num / den);
    }

    @Override public int intValue()
    {
        return (int)longValue();
    }
    @Override public long longValue()
    {
        return num / den;
    }
    @Override public float floatValue()
    {
        return (float)doubleValue();
    }
    @Override public double doubleValue()
    {
        return (double)num / (double)den;
    }

    public String toString(int radix, boolean showden1)
    {
        StringBuilder retval = new StringBuilder(Long.toString(num, radix));
        if (showden1 || den != 1L)
        {
            retval.append('/');
            retval.append(Long.toString(den, radix));
        }
        return retval.toString();
    }
    public String toString(int radix)
    {
        return toString(radix, false);
    }
    public String toString(boolean showden1)
    {
        return toString(10, showden1);
    }
    @Override public String toString()
    {
        return toString(10, false);
    }

    //
    // In the following private helper we know that the result will be
    // in lowest terms, so we can avoid going through the greatest common
    // divisor business which happens in the constructor that takes two
    // integer arguments.  This technique is also used in reciprocal(),
    // pow() and sqr().
    //
    private Rational neg()
    {
        Rational newrat = new Rational(-this.num);
        newrat.den = this.den;
        return newrat;
    }
    public Rational negate()
    {
        return num == 0L ? this : neg();
    }
    public Rational abs()
    {
        return num >= 0L ? this : neg();
    }
    public Rational reciprocal()
    {
        if (num == 0L)
        {
            throw new ArithmeticException("Reciprocal of rational zero");
        }
        //
        // We know we're in lowest terms.
        //
        Rational newrat = new Rational(this.den);
        newrat.den = this.num;
        return newrat;
    }
    public Rational min(Rational other)
    {
        return compareTo(other) <= 0 ? this : other;
    }
    public Rational max(Rational other)
    {
        return compareTo(other) >= 0 ? this : other;
    }
    public Rational copySign(Rational other)
    {
        return (this.num < 0) == (other.num < 0) ? this : negate();
    }

    //
    // In the following four arithmetic operations, we call the two-argument
    // constructor that does all the validation and normalization.
    //
    public Rational add(Rational rhs)
    {
        return new Rational(this.num * rhs.den + rhs.num * this.den,
                            this.den * rhs.den);
    }
    public Rational subtract(Rational rhs)
    {
        return new Rational(this.num * rhs.den - rhs.num * this.den,
                            this.den * rhs.den);
    }
    public Rational multiply(Rational rhs)
    {
        return new Rational(this.num * rhs.num, this.den * rhs.den);
    }
    public Rational divide(Rational rhs)
    {
        if (rhs.num == 0L)
        {
            throw new ArithmeticException("Rational division by zero");
        }
        // multiply by reciprocal of rhs
        return new Rational(this.num * rhs.den, this.den * rhs.num);
    }

    public Rational remainder(Rational divisor, RoundingMode rnd)
    {
        Rational quot = divide(divisor).round(rnd).multiply(divisor);
        return subtract(quot);
    }
    public Rational IEEEremainder(Rational divisor)
    {
        return remainder(divisor, RoundingMode.HALF_EVEN);
    }

    public Rational pow(int exp)
    {
        boolean negExp = exp < 0;
        if (negExp)
        {
            if (num == 0L)
            {
                throw new ArithmeticException(
                    "Raising Raional zero to a negative power");
            }
            exp = -exp;
        }

        //
        // We know we're in lowest terms.
        //
        Rational newrat = new Rational(1L);
        while (--exp >= 0)
        {
            newrat.num *= this.num;
            newrat.den *= this.den;
        }
        if (negExp)
        {
            long temp = newrat.num;
            newrat.num = newrat.den;
            newrat.den = temp;
        }
        return newrat;
    }
    public Rational sqr()
    {
        Rational newrat = new Rational(this.num * this.num);
        newrat.den = this.den * this.den;
        return newrat;
    }

    //
    // Two quick and dirty ways to approximate square roots
    // as examples of how to mimic other java.lang.Math methods:
    //
    public Rational sqrt()
    {
        if (signum() < 0)
        {
            throw new ArithmeticException(
                "Square root of a negative Rational");
        }
        return new Rational(Math.sqrt(doubleValue()));
    }
    public Rational sqrt(double accuracy)
    {
        if (signum() < 0)
        {
            throw new ArithmeticException(
                "Square root of a negative Rational");
        }
        // two-arg ctor validates accuracy
        return new Rational(Math.sqrt(doubleValue()), accuracy);
    }
}

// End of LRat.java
