//
// Rational.java
//
// Bill Seymour, 2025-03-17
//
// Copyright 2025 Bill Seymour.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This is an open-source rational number in Java that uses
// BigIntegers for its numerator and denominator.
//

import java.math.*;

public final class Rational extends Number implements Comparable<Rational>
{
    private BigInteger num, den;

    public static final Rational ZERO = new Rational();
    public static final Rational ONE = new Rational(BigInteger.ONE);
    public static final Rational ONE_HALF = new Rational(1L, 2L);

    //
    // The following private constructor helper is used when constructing
    // a Rational from an integer.  It's used, in some way, by every ctor.
    //
    private void fromInt(BigInteger n)
    {
        num = n;
        den = BigInteger.ONE;
    }
    public Rational()
    {
        fromInt(BigInteger.ZERO);
    }
    public Rational(long val)
    {
        fromInt(BigInteger.valueOf(val));
    }
    public Rational(BigInteger val)
    {
        fromInt(val);
    }

    //
    // Greatest common divisor, used only by Rational(long,long):
    //
    // Requires a != 0 && c > 0
    //
    private static long gcd(long a, long c)
    {
        if (a < 0L)
        {
            a = -a;
        }
        //
        // Euclid's algorithm (Knuth, vol. 2, sect. 4.5.2):
        //
        do {
            long r = a % c;
            a = c;
            c = r;
        } while (c != 0L);
        return a;
    }
    public Rational(long n, long d)
    {
        if (d == 0L)
        {
            throw new ArithmeticException(
                "Constructing Rational with zero denominator");
        }
        if (n == 0L)
        {
            fromInt(BigInteger.ZERO);
        }
        else
        {
            if (d < 0L)
            {
                n = -n;
                d = -d;
            }
            long g = gcd(n, d);
            if (g != 1L)
            {
                n /= g;
                d /= g;
            }
            num = BigInteger.valueOf(n);
            den = BigInteger.valueOf(d);
        }
    }

    //
    // The following constructor helper normalizes the fraction such that
    // the denominator is greater than zero and the fraction is reduced
    // to lowest terms.  It's shared by Rational(BigInteger,BigInteger),
    // Rational(BigDecimal,MathContext), and Rational(String,int).
    // This is basically the same logic as Rational(long,long).
    //
    private void norm(BigInteger n, BigInteger d)
    {
        int dsign = d.signum();
        if (dsign == 0)
        {
            throw new ArithmeticException(
                "Constructing Rational with zero denominator");
        }
        if (n.equals(BigInteger.ZERO))
        {
            fromInt(n);
        }
        else
        {
            if (dsign < 0)
            {
                n = n.negate();
                d = d.negate();
            }
            BigInteger g = d.gcd(n);
            if (!g.equals(BigInteger.ONE))
            {
                n = n.divide(g);
                d = d.divide(g);
            }
            num = n;
            den = d;
        }
    }
    public Rational(BigInteger n, BigInteger d)
    {
        norm(n, d);
    }

    //
    // Exact conversion of double by twidling bits:
    //
    // Assume IEEE 754 doubles:  52 significand bits,
    // 11 exponent bits excess 1023, and a sign.
    //
    // First, avoid a bunch of magic numbers:
    //
    private static final long SIGN_BIT = 0x8000_0000_0000_0000L;
    private static final long EXP_BITS = 0x7ff0_0000_0000_0000L;
    private static final long SIG_BITS = 0x000f_ffff_ffff_ffffL;
    private static final long NORM_MSB = 0x0010_0000_0000_0000L;
    private static final int SIG_BIT_COUNT = 52;
    private static final int EXP_OFFSET = 1075; // excess 1023 + 52 sig. bits
                                                // so that sig. LSB is 2**exp
    public Rational(double val)
    {
        long sig = Double.doubleToRawLongBits(val); // all bits for now

        if ((sig & ~SIGN_BIT) == 0L) // the value zero
        {
            fromInt(BigInteger.ZERO);
            return; // NB:  early return
        }

        int exp = (int)((sig & EXP_BITS) >> SIG_BIT_COUNT);
        if (exp == 0x7ff)
        {
            throw new IllegalArgumentException(
              "Can't construct rational from non-finite double");
        }

        boolean isNeg = sig < 0L;
        sig &= SIG_BITS; // now just the significand bits

        if (exp != 0) // normal value
        {
            sig |= NORM_MSB;   // assumed MSB for normals
            exp -= EXP_OFFSET; // LSB now 2**exp
        }
        else
        {
            // subnormal exponent so that LSB represents 2**-1074:
            exp = -(EXP_OFFSET - 1);
        }

        //
        // If the exponent is less than zero, we'll multiply the denominator
        // by some power of two; so to make sure that we're in lowest terms,
        // discard trailing zeros until either the significand is odd or the
        // exponent is no longer negative:
        //
        while (exp < 0 && (sig & 1L) == 0L)
        {
            sig >>= 1;
            ++exp;
        }

        //
        // Now initialize the fraction to the significand / 1:
        //
        fromInt(BigInteger.valueOf(sig));

        //
        // Adjust the value given the exponent:
        //
        if (exp > 0)
        {
            num = num.shiftLeft(exp);
        }
        else if (exp < 0)
        {
            den = den.shiftLeft(-exp);
        }

        //
        // Since either den is 1 or num is odd and den is a power of 2,
        // we know that we're in lowest terms; so all that remains is:
        //
        if (isNeg)
        {
            num = num.negate();
        }
    }

    //
    // Approximate conversion of double by continued fractions:
    //
    public Rational(double val, double err)
    {
        if (!Double.isFinite(val))
        {
            throw new IllegalArgumentException(
                "Can't construct Rational from non-finite double");
        }

        boolean isNeg = val < 0.0;
        if (isNeg)
        {
            val = -val;
        }

        if (val == 0.0)
        {
            fromInt(BigInteger.ZERO);
            return; // NB:  early return
        }

        double maxError = Math.abs(val * err);
        if (!Double.isFinite(maxError) || maxError < Double.MIN_NORMAL)
        {
            throw new ArithmeticException(
                "Non-finite or extremely small maximum error " +
                "when constructing Rational from double");
        }

        // continued fractions (Knuth, vol. 2, sect. 4.5.3):

        double val0 = val; // current fraction term

        BigInteger int0 = BigInteger.valueOf((long)val0);

        BigInteger num0 = int0,           // current approximation
                   den0 = BigInteger.ONE; // first guess:  (int)val / 1

        BigInteger num1 = BigInteger.ONE, num2;  // prev. and 2nd-prev. numer.
        BigInteger den1 = BigInteger.ZERO, den2; // prev. and 2nd-prev. denom.

        double err0 = Math.abs(val0 - int0.doubleValue());

        while (err0 > maxError)
        {
            val0 = 1.0 / (val0 - int0.doubleValue());
            if (!Double.isFinite(val0))
            {
                break; // the best we can do
            }
            int0 = BigInteger.valueOf((long)val0);

            num2 = num1;
            num1 = num0;
            num0 = int0.multiply(num1).add(num2);

            den2 = den1;
            den1 = den0;
            den0 = int0.multiply(den1).add(den2);

            double err1 = err0;
            err0 = Math.abs(val - num0.doubleValue() / den0.doubleValue());

            if (err0 >= err1) // no longer converging
            {
                break; // the best we can do
            }
        }

        //
        // Continued fractions result is known to be in lowest terms already.
        //
        num = isNeg ? num0.negate() : num0;
        den = den0;
    }

    public Rational(BigDecimal arg, MathContext mc)
    {
        BigDecimal val = arg.round(mc).stripTrailingZeros();

        BigInteger ival = val.unscaledValue();
        int sc = val.scale();
        if (sc > 0)
        {
            norm(ival, BigInteger.TEN.pow(sc));
        }
        else if (sc < 0)
        {
            fromInt(ival.multiply(BigInteger.TEN.pow(-sc)));
        }
        else // sc == 0
        {
            fromInt(ival);
        }
    }
    public Rational(BigDecimal arg)
    {
        this(arg, MathContext.UNLIMITED);
    }

    public Rational(String val, int radix)
    {
        int punctuator = val.indexOf('/');
        if (punctuator >= 0)
        {
            // numerator/denominator
            norm(new BigInteger(val.substring(0, punctuator), radix),
                 new BigInteger(val.substring(punctuator + 1), radix));
        }
        else // no slash
        {
            punctuator = val.indexOf('.');
            if (punctuator < 0)
            {
                // just an integer
                fromInt(new BigInteger(val, radix));
            }
            else
            {
                // a single value with a radix point
                StringBuilder raw = new StringBuilder(val);
                raw.deleteCharAt(punctuator);

                // unscaled value / radix**scale
                norm(new BigInteger(raw.toString(), radix),
                     BigInteger.valueOf(radix).pow(raw.length() - punctuator));
            }
        }
    }
    public Rational(String val)
    {
        this(val, 10);
    }

    @Override public int compareTo(Rational rhs)
    {
        return num.multiply(rhs.den).compareTo(den.multiply(rhs.num));
    }
    @Override public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        if (!(obj instanceof Rational))
        {
            return false;
        }
        Rational rhs = (Rational)obj;
        // We know we're in lowest terms.
        return this.num.equals(rhs.num) && this.den.equals(rhs.den);
    }
    @Override public int hashCode()
    {
        return num.hashCode() ^ den.hashCode();
    }

    public BigInteger numerator()
    {
        return num;
    }
    public BigInteger denominator()
    {
        return den;
    }
    public int signum()
    {
        return num.signum();
    }
    public Rational[] getParts()
    {
        Rational intPart  = new Rational(num.divide(den)); // rounds toward 0
        Rational fracPart = subtract(intPart);
        return new Rational[] { intPart, fracPart };
    }

    //
    // The method that actually does all the rounding:
    //
    public BigInteger toBigInteger(RoundingMode rnd)
    {
        if (den.equals(BigInteger.ONE))
        {
            return num; // It's already an integer.
        }

        //
        // Because we know we're in lowest terms, we can now infer that
        // the numerator is non-zero and is not divisible by the denominator.
        //
        BigInteger intPart = num.abs().divide(den);
        //
        // intPart is equal to the absolute value of this rounded toward zero.
        //

        boolean isNeg = num.signum() < 0; // We'll change the sign later.

        //
        // Other flags that we might not need to compute:
        //
        boolean isHalf = false;
        boolean isLess = false;
        boolean isOdd  = false;
        if (RoundingMode.HALF_UP   == rnd ||
            RoundingMode.HALF_DOWN == rnd ||
            RoundingMode.HALF_EVEN == rnd)
        {
            int halfness =
                this.abs().subtract(new Rational(intPart)).compareTo(ONE_HALF);
            isHalf = halfness == 0;
            isLess = halfness < 0;
            isOdd  = intPart.testBit(0);
        }

        switch (rnd)
        {
            case UP:
                intPart = intPart.add(BigInteger.ONE);
                break;
            case DOWN:
                // already done
                break;

            case CEILING:
                if (!isNeg)
                {
                    intPart = intPart.add(BigInteger.ONE);
                }
                break;
            case FLOOR:
                if (isNeg)
                {
                    intPart = intPart.add(BigInteger.ONE);
                }
                break;

            case HALF_UP:
                if (!isLess)
                {
                    intPart = intPart.add(BigInteger.ONE);
                }
                break;
            case HALF_DOWN:
                if (!isHalf && !isLess)
                {
                    intPart = intPart.add(BigInteger.ONE);
                }
                break;
            case HALF_EVEN:
                if (isOdd && !isLess)
                {
                    intPart = intPart.add(BigInteger.ONE);
                }
                break;

            case UNNECESSARY:
                throw new ArithmeticException("Rational.round(UNNECESSARY)");
            default:
                throw new IllegalArgumentException("Rational.round(garbage)");
        }

        return isNeg ? intPart.negate() : intPart;
    }
    public BigInteger toBigInteger()
    {
        // shorter RoundingMode.DOWN:
        return num.divide(den);
    }
    @Override public int intValue()
    {
        return toBigInteger().intValue();
    }
    @Override public long longValue()
    {
        return toBigInteger().longValue();
    }

    public BigDecimal toBigDecimal(MathContext mc)
    {
        return (new BigDecimal(num, 0)).divide(new BigDecimal(den, 0), mc);
    }
    public BigDecimal toBigDecimal()
    {
        return toBigDecimal(MathContext.DECIMAL128);
    }
    @Override public float floatValue()
    {
        return toBigDecimal().floatValue();
    }
    @Override public double doubleValue()
    {
        return toBigDecimal().doubleValue();
    }

    public String toString(int radix, boolean showden1)
    {
        StringBuilder retval = new StringBuilder(num.toString(radix));
        if (showden1 || !den.equals(BigInteger.ONE))
        {
            retval.append('/');
            retval.append(den.toString(radix));
        }
        return retval.toString();
    }
    public String toString(int radix)
    {
        return toString(radix, false);
    }
    public String toString(boolean showden1)
    {
        return toString(10, showden1);
    }
    @Override public String toString()
    {
        return toString(10, false);
    }

    //
    // In the following private helper we know that the result will be
    // in lowest terms, so we can avoid going through the greatest common
    // divisor business which happens in the constructors that take two
    // integer arguments.  This technique is also used in reciprocal(),
    // pow() and sqr().
    //
    private Rational neg()
    {
        Rational newrat = new Rational(num.negate());
        newrat.den = den;
        return newrat;
    }
    public Rational negate()
    {
        return num.equals(BigInteger.ZERO) ? this : neg();
    }
    public Rational abs()
    {
        return signum() >= 0 ? this : neg();
    }
    public Rational reciprocal()
    {
        if (num.equals(BigInteger.ZERO))
        {
            throw new ArithmeticException("Reciprocal of Rational zero");
        }
        //
        // We know we're in lowest terms.
        //
        Rational newrat = new Rational(this.den);
        newrat.den = this.num;
        return newrat;
    }
    public Rational min(Rational other)
    {
        return compareTo(other) <= 0 ? this : other;
    }
    public Rational max(Rational other)
    {
        return compareTo(other) >= 0 ? this : other;
    }
    public Rational copySign(Rational other)
    {
        return (this.signum() < 0) == (other.signum() < 0) ? this : negate();
    }

    //
    // In the following four arithmetic operations, we call constructors
    // that do all the validation and normalization.
    //
    public Rational add(Rational rhs)
    {
        return new Rational(
            this.num.multiply(rhs.den).add(rhs.num.multiply(this.den)),
            this.den.multiply(rhs.den));
    }
    public Rational subtract(Rational rhs)
    {
        return new Rational(
            this.num.multiply(rhs.den).subtract(rhs.num.multiply(this.den)),
            this.den.multiply(rhs.den));
    }
    public Rational multiply(Rational rhs)
    {
        return new Rational(this.num.multiply(rhs.num),
                            this.den.multiply(rhs.den));
    }
    public Rational divide(Rational rhs)
    {
        if (rhs.num.equals(BigInteger.ZERO))
        {
            throw new ArithmeticException("Rational division by zero");
        }
        // multiply by reciprocal of rhs
        return new Rational(this.num.multiply(rhs.den),
                            this.den.multiply(rhs.num));
    }

    public Rational round(RoundingMode rnd)
    {
        return new Rational(toBigInteger(rnd));
    }
    public Rational round()
    {
        return round(RoundingMode.HALF_UP);
    }
    public Rational ceil()
    {
        return round(RoundingMode.CEILING);
    }
    public Rational floor()
    {
        return round(RoundingMode.FLOOR);
    }
    public Rational rint()
    {
        return round(RoundingMode.HALF_EVEN);
    }
    public Rational trunc()
    {
        // shorter RoundingMode.DOWN:
        return new Rational(num.divide(den));
    }

    public Rational remainder(Rational divisor, RoundingMode rnd)
    {
        return subtract(divide(divisor).round(rnd).multiply(divisor));
    }
    public Rational IEEEremainder(Rational divisor)
    {
        return remainder(divisor, RoundingMode.HALF_EVEN);
    }

    public Rational pow(int exp)
    {
        boolean negExp = exp < 0;
        if (negExp)
        {
            if (num.equals(BigInteger.ZERO))
            {
                throw new ArithmeticException(
                    "Raising Raional zero to a negative power");
            }
            exp = -exp;
        }
        //
        // We know we're in lowest terms.
        //
        Rational newrat = new Rational();
        if (negExp)
        {
            // reciprocal
            newrat.num = this.den.pow(exp);
            newrat.den = this.num.pow(exp);
        }
        else
        {
            newrat.num = this.num.pow(exp);
            newrat.den = this.den.pow(exp);
        }
        return newrat;
    }
    public Rational sqr()
    {
        Rational newrat = new Rational();
        newrat.num = this.num.multiply(this.num);
        newrat.den = this.den.multiply(this.den);
        return newrat;
    }

    //
    // Two quick and dirty ways to approximate square roots
    // as examples of how to mimic other java.lang.Math methods:
    //
    public Rational sqrt()
    {
        if (signum() < 0)
        {
            throw new ArithmeticException(
                "Square root of a negative Rational");
        }
        return new Rational(Math.sqrt(doubleValue()));
    }
    public Rational sqrt(double accuracy)
    {
        if (signum() < 0)
        {
            throw new ArithmeticException(
                "Square root of a negative Rational");
        }
        // two-arg ctor validates accuracy
        return new Rational(Math.sqrt(doubleValue()), accuracy);
    }
}

// End of Rational.java
